package com.semaphore;

import com.sun.org.apache.xalan.internal.xsltc.cmdline.getopt.GetOpt;
import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Text;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.jdom.xpath.XPath;
import sun.misc.BASE64Encoder;

import java.io.*;
import java.net.ServerSocket;
import java.net.Socket;
import java.nio.ByteBuffer;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.logging.Logger;
import java.util.logging.LogManager;

/**
 * I KNOW.. its not elegant. It isn't meant to be.
 * Brought to you by:
 * User: semaphore
 * Date: Sep 17, 2009
 * Time: 6:46:38 PM
 */
public class TinyTSS extends Thread {
    private static Logger log = Logger.getLogger("TinyTSS");
    private static XMLOutputter xmlOut = new XMLOutputter();
    private static boolean force = false;
    private static boolean quick = false;
    private static int port = 80;

    static {
        try {
            LogManager.getLogManager().readConfiguration(TinyTSS.class.getResourceAsStream("/logger.properties"));
        } catch(Exception e) {}
        Format fmt = Format.getPrettyFormat();
        fmt.setTextMode(Format.TextMode.TRIM_FULL_WHITE);
        fmt.setIndent("\t");
        xmlOut.setFormat(fmt);
    }

    //@TODO: Allow loading multiple shsh versions for multiple ecid/version support.
    public static void main(String[] args) throws Exception {
        GetOpt g = new GetOpt(args, "f:mp:q?");
        int c;
        File mainFile = null;
        while((c = g.getNextOption()) != -1) {
            switch(c) {
                case 'f':
                    String file = g.getOptionArg();
                    mainFile = new File(file);
                    if(!mainFile.exists()) {
                        mainFile = null;
                        log.warning("That file [" + file + "] doesn't exist!");
                    }
                    break;
                case 'm':
                    force = true;
                    log.warning("Entering FORCED mode - don't blame me if you screw up your device!!!");
                    break;
                case 'q':
                    log.warning("Entering QUICK mode - TinyTSS will only reconstitute your shsh blobs from your restore binaries.");
                    quick = true;
                    break;
                case 'p':
                    try {
                        port = Integer.parseInt(g.getOptionArg());
                        log.info("Listen port set to [" + port + "]");
                    } catch(Exception e) {
                        port = 80;
                    }
                    break;
                case '?':
                default:
                    printUsage();
                    System.exit(-1);
            }
        }
        if(mainFile == null) {
            printUsage();
            System.exit(-1);
        }
        TinyTSS tss = new TinyTSS(mainFile);
        if(!quick)
            tss.start();
    }

    private static void printUsage() {
        System.err.println("\n--TinyTSS-- by semaphore v" + UmbrellaData.version + "\nshsh reconstitution info from theheadfl\n" +
                "-f - file or directory - (REQUIRED) The shsh file to load OR the saved directory you copied when you saved your 3.0 or 3.0.1 restore files.\n" +
                "-m - enable forced mode - (OPTIONAL) If the shsh blobs you saved are not complete (meaning you have less than the required 20 shsh's) this option\n" +
                "                          will force TinyTSS to keep going without throwing an exception. The only exception to this is\n" +
                "                          if you don't have your iBSS. Without that you're screwed anyway.\n" +
                "-q - enable quick mode - (OPTIONAL) If you turn on this switch, TinyTSS will not start the service on port 80. If -f points to a directory,\n" +
                "                         TinyTSS will try to reconstitute the shsh files and quit.\n" +
                "EXAMPLES:\n" +
                "\tsudo java -jar tinytss.jar -f 123123123-iphone-3.0.shsh\n" +
                "\ttinytss -f 123123123-ipt-3.1.1.shsh\n" +
                "\nBelow will check the directory listed for the shsh files and build the shsh blob file and quit.\n" +
                "\tjava -jar -f MyDirectoryWhereISaved30RestoreFiles -m -q");
    }

    private static byte[] readFile(File file) {
        try {
            FileInputStream fis = new FileInputStream(file);
            byte[] data = new byte[(int)file.length()];
            fis.read(data);
            fis.close();
            return data;
        } catch(Exception e) {
        }
        return null;
    }

    private File mainFile;
    private ServerSocket server;
    private SAXBuilder builder = new SAXBuilder(false);
    private static Map<String, Element[]> xmlCache = new HashMap<String, Element[]>();

    private TinyTSS(File firstFile) {
        this.mainFile = firstFile;
        setDaemon(false);
        setName("TinyTSS-Service");
        initServer();
        cacheFiles();
        builder.setIgnoringBoundaryWhitespace(true);
        builder.setIgnoringElementContentWhitespace(true);
        builder.setIgnoringBoundaryWhitespace(true);
        builder.setFeature("http://xml.org/sax/features/validation", false);
        builder.setFeature("http://xml.org/sax/features/use-entity-resolver2", false);
        builder.setFeature("http://apache.org/xml/features/validation/schema", false);
        builder.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd", false);
    }

    private void cacheFiles() {
        if(mainFile.isDirectory()) {
            log.info("Checking [" + mainFile + "] for SHSH blobs within a saved restore directory...");

            Map<String, String[]> rawSigs = new TreeMap<String, String[]>();

            BASE64Encoder enc = new BASE64Encoder();
            long ecid = Long.MIN_VALUE;

            File[] dirs = mainFile.listFiles(new FileFilter() {
                public boolean accept(File pathname) {
                    return pathname.isDirectory();
                }
            });
            log.info("Detecting binary restore file version...");
            BuildVersion detectedVer = null;
            for(BuildVersion bv : BuildVersion.values()) {
                List<String> bvFiles = bv.filesAsList();
                for(String fpath : bvFiles) {
                    File file = new File(mainFile, fpath);
                    if(!file.exists()) {
                        for(File dir : dirs) {
                            file = new File(dir, fpath);
                            if(file.exists()) {
                                detectedVer = bv;
                                break;
                            }
                            file = null;
                        }
                        if(file != null) {
                            detectedVer = bv;
                            break;
                        }
                    } else {
                        detectedVer = bv;
                        break;
                    }
                }
            }
            if(detectedVer == null) {
                log.severe("Unable to detect the shsh version - because you don't have the necessary .dmg files...");
                return;
            } else
                log.info("Detected [" + detectedVer.getDisplay() + "] binary restore files... attempting reconstitution!");

            Map<String, String[]> path = RequestBuilder.getFilePaths(detectedVer);
            log.info("Scanning saved directory for " + detectedVer.getDisplay() + " restore files...");
            boolean found = false;

            for(Map.Entry<String, String[]> entry : path.entrySet()) {
                String fpath = entry.getValue()[0];

                File file = new File(mainFile, fpath);
                if(!file.exists()) {
                    for(File dir : dirs) {
                        file = new File(dir, fpath);
                        if(file.exists()) {
                            break;
                        }
                        file = null;
                    }
                    if(file != null) {
                        found = true;
                        break;
                    }
                } else {
                    found = true;
                    break;
                }
            }
            if(!found) {
                log.info("Saved directory did not contain ANY " + detectedVer.getDisplay() + " restore files!");
                return;
            }

            for(Map.Entry<String, String[]> entry : path.entrySet()) {
                String key = entry.getKey();
                String fpath = entry.getValue()[0];
                if(rawSigs.containsKey(key))
                    continue;

                File file = new File(mainFile, fpath);
                if(!file.exists()) {
                    for(File dir : dirs) {
                        file = new File(dir, fpath);
                        if(file.exists())
                            break;
                        file = null;
                    }
                    if(file == null) {
                        rawSigs.clear();
                        log.severe("Unable to find:" + key + " Path:[" + fpath + "]");
                        continue;
                    }
                }
                byte[] data = readFile(file);
                ByteBuffer bb = ByteBuffer.wrap(data);
                bb.position(data.length - 2125);
                byte[] ndata = new byte[2125];
                bb.get(ndata, 0, 2125);
                data = null;
                System.gc();
                String shsh = enc.encode(ndata);
                String tag = UmbrellaData.decodeFirstTag(shsh);
                if("DICE".equals(tag)) {
                    if(ecid == Long.MIN_VALUE) {
                        ecid = UmbrellaData.decodeEcid(shsh);
                    } else {
                        long cecid = UmbrellaData.decodeEcid(shsh);
                        if(cecid != ecid) {
                            log.severe("Do NOT use files from another phone/device [" + ecid + "] != [" + cecid + "] for [" + key + "@" + fpath + "]");
                            continue;
                        }
                    }
                    rawSigs.put(key, new String[]{shsh, entry.getValue()[1], fpath});
                } else if(tag.length() == 0)
                    log.severe("Sorry this shsh is from a stock ipsw : NO ECID tag found!");
                else
                    log.severe("Sorry this shsh does not look right: Unrecognized first tag - [" + tag + "]");
                System.gc();
            }

            if(rawSigs.size() == 20) {
                log.info("Finished parsing binary restore data - found [" + rawSigs.size() + "] shsh blobs...");
            } else {
                if(!force) {
                    log.info("Unfortunately - TinyTSS was unable to find all of the files necessary to recreate your shsh blobs... use the -m flag to FORCE TinyTSS to use what it found.");
                    System.exit(-1);
                } else
                    log.severe("TinyTSS was unable to fully recreate all of your shsh blobs - don't blame me if you screw up your phone...");
            }
            reconstituteSigs(rawSigs, detectedVer, ecid);
        } else
            _cacheFiles();
    }

    private void reconstituteSigs(Map<String, String[]> rawSigs, BuildVersion cver, long ecid) {
        Document result = new Document();
        try {
            result.setDocType(new DocType("plist", "-//Apple Computer//DTD PLIST 1.0//EN", "http://www.apple.com/DTDs/PropertyList-1.0.dtd"));
            Element plist = new Element("plist");
            plist.setAttribute("version", "1.0");
            Element dict = new Element("dict");
            plist.addContent(dict);
            result.setRootElement(plist);
            for(Map.Entry<String, String[]> entry : rawSigs.entrySet()) {
                Element key = new Element("key");
                key.addContent(new Text(entry.getKey()));
                dict.addContent(key);
                Element mdict = new Element("dict");
                dict.addContent(mdict);

                key = new Element("key");
                key.addContent(new Text("Blob"));
                mdict.addContent(key);

                key = new Element("data");
                String shsh = entry.getValue()[0].trim();
                shsh = shsh.replaceAll("\\s", "");
                StringBuffer sb = new StringBuffer(shsh);
                String pad = "\n\t\t\t";
                for(int i = 0; i < sb.length(); i += 61)
                    sb.insert(i - ((i % 61) * pad.length()), pad);
                key.addContent(new Text(sb.toString() + pad));
                mdict.addContent(key);

                key = new Element("key");
                key.addContent(new Text("PartialDigest"));
                mdict.addContent(key);

                key = new Element("data");
                String pd = entry.getValue()[1].trim();
                pd = pd.replaceAll("\\s", "");
                sb = new StringBuffer(pd);
                /*
                for(int i = 0; i < sb.length(); i += 35)
                    sb.insert(i - ((i % 35) * pad.length()), pad);
                key.addContent(new Text(sb.toString() + pad));
                */
                key.addContent(new Text(sb.toString()));
                mdict.addContent(key);

                key = new Element("key");
                key.addContent(new Text("Path"));
                mdict.addContent(key);

                key = new Element("string");
                key.addContent(new Text(entry.getValue()[2]));
                mdict.addContent(key);
            }
            mainFile = new File((rawSigs.size() != 20 ? "invalid-" : "") + ecid + "-" + cver.getVersion() + ".shsh");
            UmbrellaData.backupFile(mainFile);
            xmlOut.output(result, new FileOutputStream(mainFile));
            if(!quick)
                _cacheFiles();
        } catch(Exception e) {
            e.printStackTrace();
        }
    }

    private void _cacheFiles() {
        log.info("Caching blobs...");
        if(mainFile.exists()) {
            if(mainFile.getName().equals("01.shsh")) {
                File secondFile = new File(mainFile.getParentFile(), "02.shsh");
                File thirdFile = new File(mainFile.getParentFile(), "03.shsh");
                byte[] one = readFile(mainFile);
                one = stripResult(one);
                log.info("Cached 01.shsh");
                byte[] two = readFile(secondFile);
                two = stripResult(two);
                log.info("Cached 02.shsh");
                byte[] three = readFile(thirdFile);
                three = stripResult(three);
                log.info("Cached 03.shsh");
                mergeFiles(one, two, three);
                log.info("SHSH Files have been merged into a master file -- '(ecid)-(version).shsh' -- will use this from now on. 01.shsh, 02.shsh, 03.shsh are now obsolete");
            } else {
                File masterFile = mainFile;
                byte[] master = readFile(masterFile);
                master = stripResult(master);
                log.info("Cached " + masterFile.getName());
                mergeFiles(master);
            }
//            verifyCache();
        } else {
            throw new RuntimeException("Could not find:" + mainFile.getAbsolutePath());
        }
        log.info("Finished caching blobs");
    }

    private BuildVersion verifyCache() {
        log.info("Verifying ecid shsh blobs...");
        BuildVersion foundVer = null;
        try {
            for(BuildVersion ver : BuildVersion.values()) {
                Map<String,String> digestMap = RequestBuilder.getPartialDigestMap(ver);
                boolean found = true;
                for(Map.Entry<String,Element[]> cache : xmlCache.entrySet()) {
                    String digestKey = cache.getKey();
                    Element digestElement = cache.getValue()[1];
                    XPath digests = XPath.newInstance("key[.='PartialDigest']/following-sibling::data");
                    Element ele = (Element)digests.selectSingleNode(digestElement);

                    String shshDigest = ele.getValue().replaceAll("\\s", "");
                    String plistDigest = digestMap.get(digestKey);

                    if(plistDigest == null || !shshDigest.equals(plistDigest)) {
                        found = false;
                        break;
                    }
                }
                if(found) {
                    foundVer = ver;
                    break;
                }
            }
            if(foundVer == null)
                throw new RuntimeException("Your shsh blobs are malformed... don't delete them they just need to be fixed.");
            log.info("Detected " + foundVer.getDisplay() + " shsh blobs!");
        } catch(Exception e) {
            throw new RuntimeException("Your shsh blobs are malformed... don't delete them they just need to be fixed. [" + e.getMessage() + "]");
        }
        return foundVer;
    }

    private void mergeFiles(byte[]... files) {
        xmlCache = new HashMap<String, Element[]>();
        long ecid = -1;
        BuildVersion bver = null;
        try {
            for(byte[] file : files) {
                Document doc = builder.build(new ByteArrayInputStream(file));
                XPath xp = XPath.newInstance("/plist/dict/dict/data");
                Element e = (Element)xp.selectSingleNode(doc);
                if(e != null)
                    ecid = UmbrellaData.decodeEcid(e.getValue());

                xp = XPath.newInstance("/plist/dict/dict/preceding-sibling::key[1]");
                List keysWithDicts = xp.selectNodes(doc);
                xp = XPath.newInstance("/plist/dict/key/following-sibling::dict[1]");
                List dictsOfKeys = xp.selectNodes(doc);

                for(Object key : keysWithDicts) {
                    String name = ((Element)key).getValue();
                    Element[] ele = new Element[2];
                    ele[0] = (Element)key;
                    xmlCache.put(name, ele);
                }
                int index = 0;
                for(Object dict : dictsOfKeys) {
                    String name = ((Element)keysWithDicts.get(index++)).getValue();
                    Element[] ele = xmlCache.get(name);
                    ele[1] = (Element)dict;
                }
            }
            Document doc = new Document();
            Element root = new Element("plist");
            root.setAttribute("version", "1.0");
            doc.setRootElement(root);
            Element dict = new Element("dict");
            root.addContent(dict);
            for(Map.Entry<String, Element[]> key : xmlCache.entrySet()) {
                Element[] eles = key.getValue();
                for(Element ele : eles) {
                    Element nele = (Element)ele.clone();
                    dict.addContent(nele);
                }
            }
            bver = verifyCache();

            if(files.length > 1 || mainFile.getName().startsWith("00") || mainFile.getName().startsWith("01")) {
                File master = new File(ecid + "-" + bver.getVersion() + ".shsh");
                UmbrellaData.backupFile(master);
                xmlOut.output(doc, new FileOutputStream(master));
            }
        } catch(Exception e) {
            e.printStackTrace();
        }
    }

    private byte[] stripResult(byte[] master) {
        if(new String(master).startsWith(UmbrellaData.SUCCESS_PREFIX)) {
            byte[] _master = new byte[master.length - UmbrellaData.SUCCESS_PREFIX.length()];
            System.arraycopy(master, UmbrellaData.SUCCESS_PREFIX.length(), _master, 0, _master.length);
            master = _master;
        }
        return master;
    }

    private void initServer() {
        try {
            server = new ServerSocket(port);
        } catch(IOException e) {
            System.err.println("Could not start service on port 80: Either you aren't admin or you have something running on 80!");
            System.exit(-2);
        }
    }

    public void run() {
        log.info("Starting TinyTSS service on port 80");
        while(true) {
            try {
                Socket s = server.accept();
                new Handler(s).start();
            } catch(Exception e) {
                e.printStackTrace();
                System.err.println("Something went wrong...");
                return;
            }
        }
    }

    public static class Handler extends Thread {
        private BufferedReader in;
        private DataOutputStream out;
        private SAXBuilder builder = new SAXBuilder(false);
        private boolean running = false;
        private Map<String, String> requestHeaders = new HashMap<String, String>();
        private SimpleDateFormat sdf = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss zzz");
        private static final byte[] STATUS_OK = "HTTP/1.1 200 OK\n".getBytes();

        public Handler(Socket s) {
            setDaemon(true);
            setName("TinyTSS-RequestHandler");
            builder.setIgnoringBoundaryWhitespace(true);
            builder.setIgnoringElementContentWhitespace(true);
            try {
                in = new BufferedReader(new InputStreamReader(s.getInputStream()));
                out = new DataOutputStream(s.getOutputStream());
            } catch(IOException e) {
                e.printStackTrace();
                running = false;
            }
            running = true;
        }

        private String readContent() {
            StringBuffer str = new StringBuffer();
            try {
                char[] cread = new char[8192];
                if(requestHeaders.containsKey("Transfer-Encoding") && requestHeaders.get("Transfer-Encoding").equals("chunked")) {
                    int read = 0;
                    do {
                        if(read == 0) {
                            in.mark(512);
                            String val = in.readLine();
                            if(val == null || val.length() == 0)
                                break;
                            try {
                                read = Integer.parseInt(val, 16);
                                if(read == 0)
                                    break;
                            } catch(Exception e) {
                                in.reset();
                                read = 1;
                            }
                        } else {
                            int nread = in.read(cread, 0, read);
                            if(nread != read)
                                log.severe("FAIL[" + nread + " != " + read + "]");
                            read = 0;
                            str.append(cread, 0, nread);
                        }
                    } while(true);
                } else {
                    int read;
                    do {
                        read = in.read(cread);
                        str.append(cread, 0, read);
                    } while(read != -1 && in.ready());
                }
            } catch(Exception e) {
                e.printStackTrace();
            }
            return str.toString();
        }

        private Map<String, String> readHeaders() {
            Map<String, String> headers = new HashMap<String, String>();
            try {
                String read;
                do {
                    read = in.readLine();
                    if(read.trim().indexOf(" ") > -1) {
                        String[] line = read.split(" ");
                        headers.put(line[0].replaceAll(":", ""), line[1]);
                    }
                } while(read.trim().length() != 0);
            } catch(Exception e) {
                e.printStackTrace();
            }
            return headers;
        }

        public void run() {
            if(!running)
                return;
            requestHeaders = readHeaders();
            if(requestHeaders.containsKey("GET")) {
                log.info("Received Handshake request...");
                writeHandshake();
                log.info("Wrote handshake");
            } else if(requestHeaders.containsKey("POST")) {
                log.info("Received Blob request...");
                String content = readContent();
                log.info("***Request***");
                log.info(content);
                log.info("***Request***");
                Document doc = parseRequest(content);
                writePost(doc);
                log.info("Successfully wrote blob response");
            }
        }

        private Document parseRequest(String content) {
            Document result = new Document();
            result.setDocType(new DocType("plist", "-//Apple Computer//DTD PLIST 1.0//EN", "http://www.apple.com/DTDs/PropertyList-1.0.dtd"));
            Element plist = new Element("plist");
            plist.setAttribute("version", "1.0");
            Element dict = new Element("dict");
            plist.addContent(dict);
            result.setRootElement(plist);
            try {
                Document doc = builder.build(new StringReader(content));
                XPath path = XPath.newInstance("/plist/dict/dict[data[.!='']]/preceding-sibling::key[1]");
                List nodes = path.selectNodes(doc);
                for(Object node : nodes) {
                    String key = ((Element)node).getValue();
                    if(!xmlCache.containsKey(key))
                        System.err.println("CACHE DOES NOT CONTAIN:" + key + "!!!!");
                    Element[] eles = xmlCache.get(key);
                    for(Element ele : eles) {
                        Element nele = (Element)ele.clone();
                        dict.addContent(nele);
                    }
                }
                StringWriter sw = new StringWriter();
                xmlOut.output(result, sw);
                log.info("***Response***");
                log.info(sw.toString());
                log.info("***Response***");
            } catch(Exception e) {
                e.printStackTrace();
            }
            return result;
        }

        private void writePost(Document doc) {
            try {
                StringWriter sw = new StringWriter();
                xmlOut.output(doc, sw);
                String data = UmbrellaData.SUCCESS_PREFIX + sw.toString();
                int len = data.length();
                Map<String, String> headers = new HashMap<String, String>();
                headers.put("Date", sdf.format(new Date()));
                headers.put("Server", "Apache-Coyote/1.1");
                headers.put("X-Powered-By", "Servlet 2.4; JBoss-4.0.5.GA (build: CVSTag=Branch_4_0 date=200610162339)/Tomcat-5.5");
                headers.put("Vary", "Accept-Encoding");
                headers.put("Content-Type", "text/html");
                headers.put("MS-Author-Via", "DAV");
                headers.put("Content-Length", "" + len);
                writeStatusOk();
                writeHeaders(headers);
                out.write(data.getBytes());
                out.flush();
                out.close();
            } catch(Exception e) {
            }
        }

        private void writeHandshake() {
            try {
                Map<String, String> headers = new HashMap<String, String>();
                headers.put("Date", sdf.format(new Date()));
                headers.put("Server", "Apache/2.2.11 (Debian) DAV/2 SVN/1.5.1 mod_python/3.3.1 Python/2.5.2 mod_ssl/2.2.11 OpenSSL/0.9.8g mod_transform/0.6.0");
                headers.put("Vary", "Accept-Encoding");
                headers.put("Content-Type", "text/html;charset=UTF-8");
                headers.put("Content-Length", "" + UmbrellaData.HANDSHAKE_RESPONSE_3_1.length);
                writeStatusOk();
                writeHeaders(headers);
                out.write(UmbrellaData.HANDSHAKE_RESPONSE_3_1);
                out.flush();
                out.close();
            } catch(Exception e) {
            }
        }


        private void writeHeaders(Map<String, String> headers) throws IOException {
            for(Map.Entry<String, String> entry : headers.entrySet()) {
                String line = entry.getKey() + ": " + entry.getValue() + "\n";
                out.write(line.getBytes());
            }
            out.write("\n".getBytes());
        }

        private void writeStatusOk() throws IOException {
            out.write(STATUS_OK);
        }
    }

}
